import { useContextKey } from '@kevisual/context';
import { sequelize } from '../../../modules/sequelize.ts';
import { DataTypes, Model } from 'sequelize';
import { Permission } from '@kevisual/permission';

export interface ConfigData {
  key?: string;
  version?: string;
  permission?: Permission;
}

export type Config = Partial<InstanceType<typeof ConfigModel>>;

/**
 * 用户配置
 */
export class ConfigModel extends Model {
  declare id: string;
  declare title: string;
  declare description: string;
  declare tags: string[];
  /**
   * @important 配置key， 默认可以为空，如何设置了，必须要唯一。
   */
  declare key: string;
  declare data: ConfigData; // files
  declare uid: string;
  declare hash: string;
  /**
   * 获取用户配置
   * @param key 配置key
   * @param opts 配置选项
   * @param opts.uid 用户id
   * @param opts.defaultData 默认数据
   * @returns 配置
   */
  static async getConfig(key: string, opts: { uid: string; defaultData?: any }) {
    const [config, isNew] = await ConfigModel.findOrCreate({
      where: { key, uid: opts.uid },
      defaults: {
        key,
        title: key,
        uid: opts.uid,
        data: opts?.defaultData || {},
      },
    });
    return {
      config: config,
      isNew,
    };
  }
  static async setConfig(key: string, opts: { uid: string; data: any }) {
    let config = await ConfigModel.findOne({
      where: { key, uid: opts.uid },
    });
    if (config) {
      config.data = { ...config.data, ...opts.data };
      await config.save();
    } else {
      config = await ConfigModel.create({
        title: key,
        key,
        uid: opts.uid,
        data: opts.data,
      });
    }
    return config;
  }
  /**
   * 获取上传配置
   * @param key 配置key
   * @param opts 配置选项
   * @param opts.uid 用户id
   * @returns 配置
   */
  static async getUploadConfig(opts: { uid: string }) {
    const defaultConfig = {
      key: 'upload',
      version: '1.0.0',
    };
    const config = await ConfigModel.getConfig('upload.json', {
      uid: opts.uid,
      defaultData: defaultConfig,
    });
    const data = config.config.data;
    const prefix = `/${data.key}/${data.version}`;
    return {
      config: config.config,
      isNew: config.isNew,
      prefix,
    };
  }
  static async setUploadConfig(opts: { uid: string; data: { key?: string; version?: string } }) {
    const config = await ConfigModel.setConfig('upload.json', {
      uid: opts.uid,
      data: opts.data,
    });
    return config;
  }
}
ConfigModel.init(
  {
    id: {
      type: DataTypes.UUID,
      primaryKey: true,
      defaultValue: DataTypes.UUIDV4,
      comment: 'id',
    },
    title: {
      type: DataTypes.TEXT,
      defaultValue: '',
    },
    key: {
      type: DataTypes.TEXT,
      defaultValue: '',
    },
    description: {
      type: DataTypes.TEXT,
      defaultValue: '',
    },
    tags: {
      type: DataTypes.JSONB,
      defaultValue: [],
    },
    hash: {
      type: DataTypes.TEXT,
      defaultValue: '',
    },
    data: {
      type: DataTypes.JSONB,
      defaultValue: {},
    },
    uid: {
      type: DataTypes.UUID,
      allowNull: true,
    },
  },
  {
    sequelize,
    tableName: 'kv_config',
    paranoid: true,
  },
);

ConfigModel.sync({ alter: true, logging: false }).catch((e) => {
  console.error('ConfigModel sync', e);
});

useContextKey('ConfigModel', () => ConfigModel);
